unit log.interfaces;

interface
uses
  System.Classes,
  System.SysUtils,
  localiser.interfaces;

type
  // Kind of log entry, by display importance
  Tlog_kind = (
    LK_DECORATION,                            // Decoration, separator
    LK_GROUP,                                 // group, separator
    LK_SHOWSTOPPER,                           // Show stopper, the program exit after a show stopper
    LK_CRITICAL,                              // Critcal error
    LK_EXCEPTION,                             // Exception
    LK_ERROR,                                 // error
    LK_WARNING,                               // warning
    LK_HINT,                                  // hint (light warning)
    LK_IMPORTANT,                             // important information
    LK_INFO,                                  // information
    LK_VERBOSE,                               // verbose information
    LK_VERYVERBOSE,                           // very verbose information
    LK_DEBUG                                  // debugging trace
    );


  Tlog_kinds = set of Tlog_kind;

  Tlog_kind_helper = record helper for Tlog_kind
  private
    function GetAsText: string;
    procedure SetAsKind(const Value: string);
  public
    property AsText: string read GetAsText write SetAsKind;
  end;


  // If value added please:
  // - Update Group_Kind_txt
  // - Update All getname_prefix implementations
  // - Update All getgroup_kind implementation
  Tgroup_kind = (
    GK_NONE,                 // None default value
    GK_READER,               // Root group for the stand alone reader
    GK_CONVERT_SESSION,      // Conversion group
    GK_TITLESET              // Titleset group
  );

  // specific : FFmpeg logging level
  EFFmpeg_log_level = (
    FFmpeg_log_level_auto,         // log automatique (rien si pas d'erreur)
    FFmpeg_log_level_none,         // Logge rien
    FFmpeg_log_level_error,        // log si erreur
    FFmpeg_log_level_warning,      // log si warning
    FFmpeg_log_level_info,         // log si d'info
    FFmpeg_log_level_verbose,      // log si verbeux
    FFmpeg_log_level_any           // log inconditionnel
  );



  // aliases
const
  LK_GENERAL = LK_INFO;
  LK_FATAL   = LK_CRITICAL;
  log_all_but_decoration = [
    LK_SHOWSTOPPER,                           // Show stopper, the program exit after a show stopper
    LK_CRITICAL,                              // Critcal error
    LK_EXCEPTION,                             // Exception
    LK_ERROR,                                 // error
    LK_WARNING,                               // warning
    LK_HINT,                                  // hint (light warning)
    LK_IMPORTANT,                             // important information
    LK_INFO,                                  // information
    LK_VERBOSE,                               // verbose information
    LK_VERYVERBOSE,                           // very verbose information
    LK_DEBUG                                  // debugging trace
  ];
  log_all = log_all_but_decoration + [ LK_DECORATION, LK_GROUP ];



var

  Group_Kind_txt : array[Tgroup_kind] of string =
  (
   '',
   '|GK_READER|',
   '|GK_CONVERT_SESSION|',
   '|GK_TITLESET|'
  );


{
    Format of saved log file

    Each log entry is composed of:

    information  | format              | starting column (in char.) | number of characters   | Example
    -------------------------------------------------------------------------------------------------------------------------------------------
    DATE         | YYYY-MM-DD HH:NN:SS | 1                          | 23 + Tab               | '2012-12-31 23:59:59    ' + Tab
    TYPE         | text                | 25                         | 15 + Tab               | 'Info           ' + Tab
    REPEAT COUNT | line (optionnal)    | 41 + (tab*depth)           | up to the end of line  | 'Event occured 17 times in 3.2 Sec.' + CrLf
    EVENT TEXT   | line(s)             | 41 + (tab*depth)           | up to the end of line  | 'form1 is created'


}
(*
    Log events can receive specific variables that will be evaluated when flushing

    variable           | Type                    | Description
    -------------------|-------------------------|---------------------------------------------------------------------------------------
    index              | Number                  | Index of the event
    depth              | Number                  | Parent group's depth of the event (root = 0)
    group              | Text                    | Name of parent group
    id                 | Number                  | Unique ID of the the event
    thread_id          | Number                  | ID of thread in which this event was fired
    log_type           | Text                    | Type of event (see Log_Kind_txt table)
    ticks              | Number                  | absolute tick count of the event
    ticks_relative     | Number                  | relative tick count of the event
    date               | yyyy-mm-dd              | event date, long form (ex: 2012-12-31)
    date_short         | dd mmm                  | event date, short form (ex: 31 dec.)
    time               | hh:nn:ss.zzz            | event time, with millisecond
    time_short         | hh:nn:ss                | event time, without millisecond
    time_relative      | hh:nn:ss.zzz            | event time relative to group time
    last_date          | yyyy-mm-dd              | last modification date, long form
    last_date_short    | dd mmm                  | last modification date, short form
    last_time          | hh:nn:ss.zzz            | last event modification time, with the format hh:nn:ss.zzz
    last_time_short    | hh:nn:ss                | last event modification time, with the format hh:nn:ss.zzz
    last_time_relative | hh:nn:ss.zzz            | last event modification time, with the format hh:nn:ss.zzz
    duration           | hh:nn:ss.zzz            | GROUP: Time difference between last modification of last and creation of first event
                       |                         | EVENT: Time difference between last modification and creation time
    repeat count       | Number                  | Number of times the event was repeated
    count              | Number                  | GROUP: Number of events, including subgroups
                       |                         | EVENT: Number of added lines
    text               | text                    | event text, in English
    local_text         | text                    | event text, in user locale language, otherwise English
    pid                | number                  | program id
    module             | file name & path        | module name of the process.
    context            | Text                    | if there is an active context, the prefix of the context
    child_count        | Number                  | number of child events. If this is an EVENT, it's always 0
    variant            | Text                    | The textual description of the variant stored in the event
    group_prefix       | Text                    | The prefix
    chrono_microseconds| Number                  | chronometer elapsed microseconds
    chrono_milliseconds| Number                  | chronometer elapsed milliseconds
    chrono_seconds     | Number                  | chronometer elapsed seconds
    chrono_time        | hh:nn:ss.zzz            | chronometer elapsed time with milliseconds
    chrono_time_short  | hh:nn:ss                | chronometer elapsed time without milliseconds
    chrono_percent     | Number                  | chronometer percentage compared to group
*)


type
   Log_info_item = (
     lii_index,
     lii_depth,
     lii_group,
     lii_id,
     lii_thread_id,
     lii_type,
     lii_ticks,
     lii_ticks_relative,
     lii_date,
     lii_date_short,
     lii_time,
     lii_time_short,
     lii_time_relative,
     lii_last_date,
     lii_last_date_short,
     lii_last_time,
     lii_last_time_short,
     lii_last_time_relative,
     lii_duration,
     lii_repeat_count,
     lii_add_count,
     lii_text,
     lii_pid,
     lii_module,
     lii_context,
     lii_child_count,
     lii_variant,
     lii_group_prefix,
     lii_chrono_microseconds,
     lii_chrono_milliseconds,
     lii_chrono_seconds,
     lii_chrono_time,
     lii_chrono_time_short,
     lii_chrono_percent,
     nb_info_items);

   log_info_items = set of Log_info_item;

   log_item_fmt = record
     name: string;
     width: integer;
   end;

var
  log_info_table: array[Log_info_item] of log_item_fmt = (
     ( name: 'index';                  width: 06; ),     // number (0..999999)
     ( name: 'depth';                  width: 02; ),     // number (0..99, more is not reasonable)
     ( name: 'group';                  width: 12; ),     // string
     ( name: 'id';                     width: 06; ),     // number (0..999999)
     ( name: 'thread_id';              width: 04; ),     // number (0..FFFF)
     ( name: 'type';                   width: 12; ),     // string
     ( name: 'ticks';                  width: 19; ),     // number (int64)
     ( name: 'ticks_relative';         width: 19; ),     // number (int64)
     ( name: 'date';                   width: 10; ),     // yyyy-mm-dd
     ( name: 'date_short';             width: 07; ),     // dd mmm
     ( name: 'time';                   width: 12; ),     // hh:nn:ss.zzz
     ( name: 'time_short';             width: 08; ),     // hh:nn:ss
     ( name: 'time_relative';          width: 12; ),     // hh:nn:ss.zzz
     ( name: 'last_date';              width: 10; ),     // yyyy-mm-dd
     ( name: 'last_date_short';        width: 06; ),     // dd mmm
     ( name: 'last_time';              width: 12; ),     // hh:nn:ss.zzz
     ( name: 'last_time_short';        width: 08; ),     // hh:nn:ss
     ( name: 'last_time_relative';     width: 12; ),     // hh:nn:ss.zzz
     ( name: 'duration';               width: 00; ),     // explicit duration string
     ( name: 'repeat_count';           width: 06; ),     // number (0..999999)
     ( name: 'add_count';              width: 06; ),     // number (0..999999)
     ( name: 'text';                   width: 00; ),     // string
     ( name: 'pid';                    width: 04; ),     // number (0..FFFF)
     ( name: 'module';                 width: 00; ),     // string
     ( name: 'context';                width: 00; ),     // context prefix
     ( name: 'child_count';            width: 06; ),     // number (0..999999)
     ( name: 'variant';                width: 00; ),     // variant
     ( name: 'group_prefix';           width: 00; ),     // variant
     ( name: 'chrono_microseconds';    width: 10; ),     // number (0..9999999999)
     ( name: 'chrono_milliseconds';    width: 07; ),     // number (0..9999999)
     ( name: 'chrono_seconds';         width: 04; ),     // number (0..9999)
     ( name: 'chrono_time';            width: 12; ),     // hh:nn:ss.zzz
     ( name: 'chrono_time_short';      width: 08; ),     // hh:nn:ss
     ( name: 'chrono_percent';         width: 06; ),     // xx.xx%
     ( name: '';                       width: 00; )
  );

type
    log_event_info = (
     lei_type,
     lei_id,
     lei_group_id,
     lei_index,
     lei_depth,
     lei_thread_id,
     lei_ticks,
     lei_when,
     lei_last_ticks,
     lei_last_when,
     lei_add_count,
     lei_repeat_count,
     lei_text,
     lei_local_text,
     lei_pp_format,
     lei_group_name,
     lei_group_header,
     lei_group_footer,
     lei_root_values
    );

    log_event_infos = set of log_event_info;

type
    ELogLoadType = (
      lltUnknwon,
      lltNew,
      lltLegacy,
      lltMultiple,
      lltError,
      lltLegacyError
    );


type
  ILogStorage            = interface;
  ILogStorageList        = interface;
  ILogStorageEnumerator  = interface;
  ILogGroup           = interface;
  ILogGroupList       = interface;
  ILogGroupEnumerator = interface;
  ILogEvent           = interface;
  ILogEventList       = interface;
  ILogEventEnumerator = interface;

  TLogEntriesSelection = array of ILogEvent;

  TOnLogeventAdded = procedure(Sender: TObject; entry: ILogEvent) of object;
  TOnSelectLogEvent = procedure(Sender: TObject; entry: ILogEvent; var do_include, do_stop: boolean) of object;
//  // Add a formated log entry
  TLogAddFormatedEvent  = procedure(Sender: TObject;
                                     Kind: TLog_Kind;
                                     Const Text: string;
                                     Const Params: Array of const) of object;

  TOnLogFileLoaded = procedure(Sender: TObject; LogGroup: ILogGroup; LoadingType: ELogLoadType) of object;

  TLogLocaliseProc = function (original: string) : string;

  { A Storage represent a place where the log is saved, in which occasion and what is included }
  ILogStorage = interface
    ['{7A14367C-0995-44C2-9656-09882300959B}']
  // props accessors
    function getcrash_log: boolean;
    function getdecoration_padding: integer;
    function Getdecoration_pattern: string;
    function getdiscard: Tlog_kinds;
    function Getenabled: boolean;
    function Getevent_header: string;
    function getfile_name: string;
    function getflush_interval: double;
    function getflush_on: Tlog_kinds;
    function getfq_filename: string;
    function Getgroup: ILogGroup;
    function Getgroup_indent: integer;
    function getmost_recent_first: boolean;
    function Geton_select: TOnSelectLogEvent;
    function getuse_locale: boolean;
    procedure setcrash_log(const Value: boolean);
    procedure setdecoration_padding(const Value: integer);
    procedure Setdecoration_pattern(const Value: string);
    procedure setdiscard(const Value: Tlog_kinds);
    procedure Setenabled(const Value: boolean);
    procedure Setevent_header(const Value: string);
    procedure setfile_name(const Value: string);
    procedure setflush_interval(const Value: double);
    procedure setflush_on(const Value: Tlog_kinds);
    procedure Setgroup(const Value: ILogGroup);
    procedure Setgroup_indent(const Value: integer);
    procedure setmost_recent_first(const Value: boolean);
    procedure Seton_select(const Value: TOnSelectLogEvent);
    procedure setuse_locale(const Value: boolean);
    procedure setfq_filename(const Value: string);


    function do_flush(entry: ILogEvent = nil): boolean;
    procedure Assign(src: ILogStorage);
    procedure preset_crashlog(aname: string = 'crash');


    // fully qualified filename + path
    property fq_filename: string read getfq_filename write setfq_filename;
    // owning group for this storage
    property group: ILogGroup read Getgroup write Setgroup;
    // filename (only) of the log file
    property file_name: string read getfile_name write setfile_name;
    // sort entries with the most recent first
    property most_recent_first: boolean read getmost_recent_first write setmost_recent_first;
    // time, in seconds, between each write of the log. 0 = not written at interval
    property flush_interval: double read getflush_interval write setflush_interval;
    // flushes log if any of the set is added
    property flush_on: Tlog_kinds read getflush_on write setflush_on;
    // filter which entries that shall be added
    property discard: Tlog_kinds read getdiscard write setdiscard;
    // locale or english log entry shall be written
    property use_locale: boolean read getuse_locale write setuse_locale;
    // is it a crash log file
    property crash_log: boolean read getcrash_log write setcrash_log;
    // min number of character used in decoration / header / footer lines
    property decoration_padding: integer read getdecoration_padding write setdecoration_padding;
    // symbol used to pad decoration / header / footer lines
    property decoration_pattern: string read Getdecoration_pattern write Setdecoration_pattern;
    // how event headers are rendered in log file
    property event_header: string read Getevent_header write Setevent_header;
    // group indentation, in number of characters
    property group_indent: integer read Getgroup_indent write Setgroup_indent;
    // enable or not the storage. If disabled, all files are deleted and nothing is saved
    property enabled: boolean read Getenabled write Setenabled;

  end;

  ILogFileInternal = interface
    ['{9A9B2EE2-36C1-44B1-A95C-C59F83C35944}']
    procedure notify_storage(event: ILogEvent);
  end;

  ILogStorageList = interface
    ['{3DD439BD-A82C-4372-B490-B29A682AECB8}']
    function GetCount: Integer;
    function GetItems(Index: Integer): ILogStorage;

    function add(const item: ILogStorage): integer;
    function remove(const item: ILogStorage): integer;
    function IndexOf(const Item: ILogStorage): Integer;
    function first: ILogStorage;
    function last: ILogStorage;
    function GetEnumerator: ILogStorageEnumerator;
    procedure clear;
    property Count: Integer read GetCount;
    property Items[Index: Integer]: ILogStorage read GetItems; default;
  end;

  { A Log context represent an opaque data which is associated to a given log group, or prefix.
    It allow to link some generic log (such as FFMPEG log) to entries, and operate a selection on this }
  ILogContext = interface
    ['{5C13EA88-E274-4EFB-960A-A143B40EAA0F}']
    function Getcategory: integer;
    function Getdata: pointer;
    function Getgroup: ILogGroup;
    function Getname: string;
    function Getprefix: string;
    function Getthread_id: NativeUint;
    procedure Setcategory(const Value: integer);
    procedure Setdata(const Value: pointer);
    procedure Setgroup(const Value: ILogGroup);
    procedure Setname(const Value: string);
    procedure Setprefix(const Value: string);
    procedure Setthread_id(const Value: NativeUint);

    property name: string read Getname write Setname;
    property prefix: string read Getprefix write Setprefix;
    property group: ILogGroup read Getgroup write Setgroup;
    property category: integer read Getcategory write Setcategory;
    property data: pointer read Getdata write Setdata;
    property thread_id: NativeUint read Getthread_id write Setthread_id;
  end;

  ILogcontextEnumerator = interface
    ['{F2F4C8E5-C7A1-4054-BA3F-3A06980298B1}']
    function GetCurrent: ILogcontext;
    function MoveNext: Boolean;
    property Current: ILogcontext read GetCurrent;
  end;

  ILogContextList = interface
    ['{5178DB88-76ED-482C-AE43-356D65F2FE0D}']
    function GetCount: Integer;
    function GetItems(Index: Integer): ILogcontext;

    function add(const item: ILogcontext): integer;
    function remove(const item: ILogcontext): integer;
    function IndexOf(const Item: ILogcontext): Integer;
    function first: ILogcontext;
    function last: ILogcontext;
    function GetEnumerator: ILogcontextEnumerator;
    procedure clear;
    function find_by_data(adata: pointer; var ctx: ILogContext): boolean;
    property Count: Integer read GetCount;
    property Items[Index: Integer]: ILogcontext read GetItems; default;
  end;

  ILogEvent = interface
    ['{74C842F7-936D-440D-AFE8-88E5EE2C3546}']
    function getadd_count: integer;
    function Getchild_count: cardinal;
    function getcommit_id: cardinal;
    function getdata: Pointer;
    function getdepth: integer;
    function Getid: cardinal;
    function Getindex: integer;
    function getlast_ticks: Int64;
    function getlast_when: TDateTime;
    function getlast_when_relative: TDateTime;
    function getlocal_or_native_text: string;
    function Getlocal_text: string;
    function Getlog_type: TLog_Kind;
    function Getlog_type_astext: string;
    function Getparent: ILogGroup;
    function Getparent_group_name: string;
    function Getparent_group_prefix: string;
    function GetPreProcessFormat: string;
    function getrepeat_count: Longword;
    function Gettext: string;
    function Getthread_id: NativeUInt;
    function Getticks: Int64;
    function Getticks_relative: int64;
    function getupdating: boolean;
    function Getvalue: variant;
    function getvalue_as_text: string;
    function Getwhen: TDateTime;
    function getwhen_relative: TDateTime;
    function getchrono_elapsed_ticks: int64;
    function getchrono_percent: double;
    procedure register_localise_proc(aProc : TLogLocaliseProc);
    function localise(const original: string): string;
    procedure setdata(const Value: Pointer);
    procedure Setlocal_text(const Value: string);
    procedure SetPreProcessFormat(const Value: string);
    procedure Settext(const Value: string);
    procedure Setvalue(const aValue: variant);

    function preprocess(const txt: string; locale: boolean = true; use_width: boolean = false): string;

    procedure add_commandline(hide_if_contains: string = '');
    procedure add_cpu;
    procedure add_exception(Exception_info: pointer);
    procedure add_memory;
    procedure add_startup_info;
    procedure add_winerror;

    // add a static text
    procedure add(Const txt: string; aContext: ILogContext = nil; do_localise: boolean = true); overload;
    // add a formatted text (support standard or ordered format)
    procedure add(Const fmt: string; Const Params: Array of const; aContext: ILogContext = nil; do_localise: boolean = true); overload;
    // add a text format, which will include representation of a value ex: add('{value} skip root menu',false);
    procedure add(const fmt: string; Const aValue: boolean; aContext: ILogContext = nil; do_localise: boolean = true); overload;
//    // add a child event representing a flag or a boolean
//    procedure add_boolean(Const event_text: string; Value: Boolean; aContext: ILogContext = nil; do_localise: boolean = true); overload;
//    // add a child event with a composed text
//    procedure add_composed(Const fmt: string; Const Params: Array of string; aContext: ILogContext = nil; do_localise: boolean = true); overload;


    // Chronometric functions.
    // Start chronometer
    procedure chrono_start;
    // Stop chronometer
    procedure chrono_stop;
    // Reset chronometer (elapsed ticks return to 0)
    procedure chrono_reset;


    function BeginUpdate: boolean;
    function EndUpdate: boolean;
    function check_repeat(event: ILogEvent): boolean;
    function is_child_of(log_group: ILogGroup): Boolean;

    // set the repeat count and the Tick count of the last repeat
    procedure set_repeat(anb_repeat: Integer; alast_tick: int64);


    // parent event (can be nil)
    property parent: ILogGroup read Getparent;
    // name of parent group (it's a helper)
    property parent_group_name: string read Getparent_group_name;
    // prefix of parent group (it's a helper)
    property parent_group_prefix: string read Getparent_group_prefix;
    // index in parent group
    property index: integer read Getindex;
    // depht (number of parent groups) of this event
    property depth: integer read getdepth;
    // event unique ID (systemwide), in fact index of occurence
    property id: cardinal read Getid;
    // commit id, used in storages to find out if something has changed and need to be flushed
    property commit_id:cardinal read getcommit_id;
    // number of child events (0 if it's a simple event)
    property child_count: cardinal read Getchild_count;
    // thread ID in which the event was created
    property thread_id: NativeUInt read Getthread_id;
    // Type of log event
    property log_type: TLog_Kind read Getlog_type;
    // Textual description of log event type
    property log_type_as_text: string read Getlog_type_astext;
    // tick count (perf. counter)
    property ticks: Int64 read Getticks;
    // tick count (perf. counter), relative to parent event
    property ticks_relative: int64 read Getticks_relative;
    // absolute date/time of the log event
    property when: TDateTime read Getwhen;
    // relative date/time of the log event
    property when_relative: TDateTime read getwhen_relative;
    // absolute date/time of the last modification
    property last_when: TDateTime read getlast_when;
    // relative date/time of the last modification
    property last_when_relative: TDateTime read getlast_when_relative;
    // relative tick count of the last modification
    property last_ticks: Int64 read getlast_ticks;
    // Number of additions in the event (= number of lines in the log)
    property add_count: integer read getadd_count;
    // Number of time the same event was repeated
    property repeat_count: Longword read getrepeat_count;
    // Chronometric elapsed tick count
    property chrono_elapsed_ticks: int64 read getchrono_elapsed_ticks;
    // Chronometric percentage [0..100], compared to parent total elapsed ticks
    property chrono_percent: double read getchrono_percent;
    // user defined data, such as VTV node
    property data: Pointer read getdata write setdata;
    // a value, any type
    property value: variant read Getvalue write Setvalue;
    // textual representation of the value
    property value_as_text: string read getvalue_as_text;
    // The event text, in native language
    property text: string read Gettext write Settext;
    // The event text, in locale langage, can be empty if only native language is available
    property local_text: string read Getlocal_text write Setlocal_text;
    // The text, in locale langage or if not available, in native language
    property local_or_native_text: string read getlocal_or_native_text;
    // Prepocess format empty if default used
    property PreProcessFormat: string read GetPreProcessFormat write SetPreProcessFormat;
     // True when in updating mode
    property updating: boolean read getupdating;
  end;

  ILogEventEnumerator = interface
    ['{EF6D49E8-B402-4FF0-973A-90D7270EFE32}']
    function GetCurrent: ILogEvent;

    function MoveNext: Boolean;
    property Current: ILogEvent read GetCurrent;
  end;


  ILogEventList = interface
    ['{0C2E41C4-7020-4ECD-8760-5913568A7FF7}']
    function GetCount: Integer;
    function GetItems(Index: Integer): ILogEvent;

    function GetEnumerator: ILogEventEnumerator;
    procedure clear;
    function addentry(const item: ILogEvent): integer;
    function first: ILogEvent;
    function last: ILogEvent;
    property Count: Integer read GetCount;
    property Items[Index: Integer]: ILogEvent read GetItems; default;
  end;


  ILogGroup = interface(ILogEvent)
  ['{10BE6743-E350-4502-AE18-6D5146213CC7}']
    function getdiscard: Tlog_kinds;
    function getevent(index: integer): ILogEvent;
    function getevent_count: integer;
    function getevent_first: ILogEvent;
    function getevent_last: ILogEvent;
    function getevent_list: ILogEventList;
    function getfooter: string;
    function getgroup_list: ILogGroupList;
    function getgroup_king: Tgroup_kind;
    function getheader: string;
    function Getname: string;
    function getname_prefix: string;
    function GetOnAdded: TOnLogeventAdded;
    function GetOnChanged: TNotifyEvent;
    function getstorage(index: integer): ILogStorage;
    function getstorage_count: integer;
    function getstorage_first: ILogStorage;
    function getstorage_last: ILogStorage;
    function getstorage_list: ILogStorageList;
    procedure setdiscard(const Value: Tlog_kinds);
    procedure setfooter(const Value: string);
    procedure setheader(const Value: string);
    procedure Setname(const Value: string);
    procedure SetOnAdded(const Value: TOnLogeventAdded);
    procedure SetOnChanged(const Value: TNotifyEvent);
    function Getautostart: boolean;
    procedure Setautostart(const Value: boolean);
    function getstarted: boolean;
    function getsaved_on: Tlog_kinds;
    procedure setsaved_on(const Value: Tlog_kinds);
    function Getgroup_tag: cardinal;
    procedure setgroup_tag(const Value: cardinal);


    function GetEnumerator: ILogEventEnumerator;

    // clear all entries and child groups
    procedure clear_entries;
    // clear all storage
    procedure clear_storages;
    // clear everything
    procedure clear;

    procedure remove;

    // add a log entry
    procedure AddLog(alog_type: Tlog_kind; Const txt: string; aContext: ILogContext = nil; do_localise: boolean = true); overload;
    // add a formatted text (support standard or ordered format)
    procedure AddLog(alog_type: Tlog_kind; Const fmt: string; Const Params: Array of const; aContext: ILogContext = nil; do_localise: boolean = true); overload;
    // add a text format, which will include representation of a value ex: add('{value} skip root menu',false);
    procedure AddLog(alog_type: Tlog_kind; const fmt: string; Const aValue: boolean; aContext: ILogContext = nil; do_localise: boolean = true); overload;
    // add a log entry at a given timestamp
    procedure AddLogDefered(alog_type: Tlog_kind; Const txt: string; aTimeStamp: Int64 = 0; aThreadID: integer = 0; aRepeat: integer = 0; aEndTimeStamp: int64 = 0);

//    procedure add(Kind: TLog_Kind; Const event_text: string; aContext: ILogContext = nil; do_localise: boolean = true); overload;
//    // add a child event with a formatted text
//    procedure add(Kind: TLog_Kind; Const fmt: string; Const Params: Array of const; aContext: ILogContext = nil; do_localise: boolean = true); overload;
//    // add a child event representing a flag or a boolean
//    procedure add_boolean(Kind: TLog_Kind; Const event_text: string; Value: Boolean; aContext: ILogContext = nil; do_localise: boolean = true); overload;
//    // add a child event with a composed text
//    procedure add_composed(Kind: TLog_Kind; Const fmt: string; Const Params: Array of string; aContext: ILogContext = nil; do_localise: boolean = true); overload;
//


    // find a child group
    function find_or_add_group(const name: string; var group: ILogGroup): boolean;
    function find_group_of_thread(Id: NativeUint; var group: ILogGroup): boolean;
    function add_exception_context(txt: string; E: Exception): ILogEvent;

    // select a list of log events with criterias
    function select(list: ILogEventList; sel_discard: Tlog_kinds = []; stop_at: ILogEvent = nil): boolean;
    // return TRUE if log group contains at least one entry in saved_on
    function is_saved: boolean;

    // create a storage without registering it
    function new_storage: ILogStorage;
    // register a given storage to this group
    function register_storage(new_storage: ILogStorage): integer;

    procedure start;
    procedure stop;

    procedure enable_storages(enabled: boolean);

    procedure lock_parent_notify;
    procedure unlock_parent_notify;


    // Create a new empty event
    function new_event(kind: Tlog_kind): ILogEvent;
    // Create a new empty group
    function new_group(const aName: string): ILogGroup;

    // adds a random group with it's name seeded,
    // to emulate add_random_group do not fill the second parameter
    function add_seeded_group(var group: ILogGroup; seed: Tgroup_kind = GK_NONE): boolean;


    // Loads log from file
    // If crashlist file is selected, LoadFromFiles will be called with all files inside crash list
    function LoadFromFile(const FileName: string): boolean;
    function LoadFromStream(const FileStream: TStream; const aCaption: string): boolean;

    property autostart: boolean read Getautostart write Setautostart;
    // number of event entries
    property event_count: integer read getevent_count;
    // accessor to event event
    property event[index: integer]: ILogEvent read getevent;
    // list of event
    property event_list: ILogEventList read getevent_list;
    // first event event
    property event_first: ILogEvent read getevent_first;
    // last event event
    property event_last: ILogEvent read getevent_last;
    // number of storage
    property storage_count: integer read getstorage_count;
    // accessor to storage
    property storage[index: integer]:ILogStorage read getstorage;
    // list of storage
    property storage_list: ILogStorageList read getstorage_list;
    // first storage event
    property storage_first: ILogStorage read getstorage_first;
    // last storage event
    property storage_last: ILogStorage read getstorage_last;
    // list of child groups
    property group_list: ILogGroupList read getgroup_list;
    // what entries shall be discarded
    property discard: Tlog_kinds read getdiscard write setdiscard;
    // what entries trigger saving of this log group
    property saved_on:Tlog_kinds read getsaved_on write setsaved_on;
    // group name, must not contains dot
    property name: string read getname write setname;
    // group name prefix, returns empty value if unknown prefix
    property name_prefix: string read getname_prefix;
    // returns grou kind extracted from group name
    property group_kind: Tgroup_kind read getgroup_king;
    property group_tag: cardinal read Getgroup_tag write setgroup_tag;

    // group header, for log presentation
    property header: string read getheader write setheader;
    // group footer, for log presentation
    property footer: string read getfooter write setfooter;

    property started: boolean read getstarted;

    // Event called each time a event event is added
    property OnAdded: TOnLogEventAdded read GetOnAdded write SetOnAdded;
    // Event called when at least 1 event as been added, and when we are not in updating mode
    property OnChanged: TNotifyEvent read GetOnChanged write SetOnChanged;
  end;

  ILogGroupInternal = interface(ILogGroup)
    ['{67E0930D-235F-49FD-84C6-81CEDAC4F36B}']
    procedure notify_parent(event: ILogEvent);
    procedure notify_change(event: ILogEvent);
  end;

  ILogStorageEnumerator = interface
    ['{F2F4C8E5-C7A1-4054-BA3F-3A06980298B1}']
    function GetCurrent: ILogStorage;
    function MoveNext: Boolean;
    property Current: ILogStorage read GetCurrent;
  end;

  ILogGroupEnumerator = interface
  ['{69F2C7D1-8DE2-430E-9BD6-101A65B3C0A9}']
    function GetCurrent: ILogGroup;
    function MoveNext: Boolean;
    property Current: ILogGroup read GetCurrent;
  end;

  ILogGroupList = interface
    ['{DFB6E42F-0B3B-4259-B85D-5F94B821C457}']
    function GetCount: Integer;
    function GetItems(Index: Integer): ILogGroup;

    function GetEnumerator: ILogGroupEnumerator;
    procedure clear;
    function add(const item: ILogGroup): integer;
    function first: ILogGroup;
    function last: ILogGroup;
    property Count: Integer read GetCount;
    property Items[Index: Integer]: ILogGroup read GetItems; default;
  end;

  ILogRoot = interface(ILogGroup)
    ['{A4DA70DE-C2C4-4ED5-9E70-6DBD0E80143A}']
    function Getcontext_list: ILogContextList;
    function Getcrashlist_name: string;
    function Getfalse_string: string;
    function getfq_filename: string;
    function getfq_path: string;
//    function Getlocation: Tknown_folder;
    function Getlong_date_format: string;
    function Getlong_time_format: string;
    function Getnative_language: string;
    function GetOnLogFileLoaded: TOnLogFileLoaded;
    function Getpath: string;
    function Getshort_date_format: string;
    function Getshort_time_format: string;
    function Gettrue_string: string;
    procedure Setcrashlist_name(const Value: string);
    procedure Setfalse_string(const Value: string);
//    procedure Setlocation(const Value: Tknown_folder);
    procedure Setlong_date_format(const Value: string);
    procedure Setlong_time_format(const Value: string);
    procedure Setnative_language(const Value: string);
    procedure SetOnLogFileLoaded(const Value: TOnLogFileLoaded);
    procedure Setpath(const Value: string);
    procedure Setshort_date_format(const Value: string);
    procedure Setshort_time_format(const Value: string);
    procedure Settrue_string(const Value: string);

    function register_crash(src: ILogStorage): boolean;
    function unregister_crash(src: ILogStorage): boolean;
    function sync(src: ILogStorage; adding: boolean): boolean;
    function build_crash_list(list: TStringlist): integer;

    function find_context(out ctx: ILogContext; data: Pointer; can_create: boolean = true): boolean;
    function remove_context(ctx: ILogContext): boolean;

    procedure clear_contexts;

    property fq_filename: string read getfq_filename;
    property fq_path: string read getfq_path;

    property context_list: ILogContextList read Getcontext_list;
//    property location: Tknown_folder read Getlocation write Setlocation;
    property crashlist_name: string read Getcrashlist_name write Setcrashlist_name;
    property path: string read Getpath write Setpath;
    property true_string: string read Gettrue_string write Settrue_string;
    property false_string: string read Getfalse_string write Setfalse_string;
    property short_date_format: string read Getshort_date_format write Setshort_date_format;
    property long_date_format: string read Getlong_date_format write Setlong_date_format;
    property short_time_format: string read Getshort_time_format write Setshort_time_format;
    property long_time_format: string read Getlong_time_format write Setlong_time_format;
    property native_language: string read Getnative_language write Setnative_language;
    property OnLogFileLoaded: TOnLogFileLoaded read GetOnLogFileLoaded write SetOnLogFileLoaded;
  end;


// Add a log entry that will not be localised (translated)
procedure AddUnlocLog(Kind: TLog_Kind; Const Text: string);

// Log some text in the log group associated with the calling thread. If there is no log group associated, log into root
procedure log_thread_context(Kind: TLog_Kind; Const Text: string); overload;

// Log some text in the log group associated with the calling thread. If there is no log group associated, log into root
procedure log_thread_context(Kind: TLog_Kind; Const fmt: string; Const Params: Array of const); overload;

// create a log event from the exception context, and add it to the log root
function Log_exception_context(txt: string; E: Exception): ILogEvent;

// add a log entry to the log root
procedure AddLog(Kind: TLog_Kind; Const Text: string); overload;

// same as above, with a formatted text
procedure AddLog(Kind: TLog_Kind; Const fmt: string; Const Params: Array of const); overload;

// add a log entry for a switch value
procedure AddLog(Kind: TLog_Kind; Const Text: string; Value: Boolean); overload;

// Global AddLog with composed string
//procedure AddComposedLog(Kind: TLog_Kind; Const fmt: string; Const Params: Array of string); overload;

// add a log entry to a given log group
procedure AddLog(const group_name: string; Kind: TLog_Kind; Const Text: string); overload;

// same as above, with a formatted text
procedure AddLog(const group_name: string; Kind: TLog_Kind; Const fmt: string; Const Params: Array of const); overload;

// add a log entry in a log group, for a switch value
procedure AddLog(const group_name: string; Kind: TLog_Kind; Const Text: string; Value: Boolean); overload;

// Global AddLog with composed string
//procedure AddComposedLog(const group_name: string; Kind: TLog_Kind; Const fmt: string; Const Params: Array of string); overload;

// add a log entry to a given log group
procedure AddLog(group: ILogGroup; Kind: TLog_Kind; Const Text: string); overload;

// same as above, with a formatted text
procedure AddLog(group: ILogGroup; Kind: TLog_Kind; Const fmt: string; Const Params: Array of const); overload;

// add a log entry for a switch value
procedure AddLog(group: ILogGroup; Kind: TLog_Kind; Const Text: string; Value: Boolean); overload;

// Add a log group
function AddLogGroup(const name: string; group: ILogGroup = nil): ILogGroup;

// The root of the logs
var
  log_root: ILogRoot;

implementation
uses debug_helper;

procedure log_thread_context(Kind: TLog_Kind; Const Text: string); overload;
var
  Grp: ILogGroup;
begin

  if Assigned(log_root) then
  begin
    if log_root.find_group_of_thread(TThread.CurrentThread.ThreadID,Grp) then
    begin
      grp.AddLog(Kind,Text);
    end else begin
      // what do ???
      log_root.AddLog(Kind,Text);
    end;
  end;
end;

procedure log_thread_context(Kind: TLog_Kind; Const fmt: string; Const Params: Array of const); overload;
var
  Grp: ILogGroup;
begin

  if Assigned(log_root) then
  begin
    if log_root.find_group_of_thread(TThread.CurrentThread.ThreadID,Grp) then
    begin
      grp.AddLog(Kind,fmt,params);
    end else begin
      // what do ???
      log_root.AddLog(Kind,fmt,params);
    end;
  end;
end;

function Log_exception_context(txt: string; E: Exception): ILogEvent;
begin
  if assigned(log_root) then
    result := log_root.add_exception_context(txt,E)
  else
    result := nil;
end;

procedure AddUnlocLog(Kind: TLog_Kind; Const Text: string);
begin
  if Assigned(log_root) then
    log_root.AddLog(Kind,Text)
  else begin
    // void
  end;
end;

// add a log entry to the log root
procedure AddLog(Kind: TLog_Kind; Const Text: string); overload;
begin
  if Assigned(log_root) then
    log_root.AddLog(Kind,Text)
  else begin

  end;
end;

// same as above, with a formatted text
procedure AddLog(Kind: TLog_Kind; Const fmt: string; Const Params: Array of const); overload;
begin
  if Assigned(log_root) then
    log_root.AddLog(Kind,fmt,Params)
  else begin

  end;
end;

// add a log entry for a switch value
procedure AddLog(Kind: TLog_Kind; Const Text: string; Value: Boolean); overload;
begin
  if Assigned(log_root) then
    log_root.AddLog(Kind,Text,Value)
  else begin

  end;
end;

// Global AddLog with composed string
procedure AddComposedLog(Kind: TLog_Kind; Const fmt: string; Const Params: Array of const); overload;
begin
  if Assigned(log_root) then
    log_root.AddLog(Kind,fmt,Params)
  else begin

  end;
end;

// add a log entry to a given log group
procedure AddLog(const group_name: string; Kind: TLog_Kind; Const Text: string); overload;
var
  lg: ILogGroup;
begin
  if Assigned(log_root) then
  begin
    log_root.find_or_add_group(group_name,lg);
    lg.AddLog(Kind,Text);
  end else begin

  end;
end;

// same as above, with a formatted text
procedure AddLog(const group_name: string; Kind: TLog_Kind; Const fmt: string; Const Params: Array of const); overload;
var
  lg: ILogGroup;
begin
  if Assigned(log_root)  then
  begin
    log_root.find_or_add_group(group_name,lg);
    lg.AddLog(Kind,fmt,Params);
  end else begin

  end;
end;

// add a log entry for a switch value
procedure AddLog(const group_name: string; Kind: TLog_Kind; Const Text: string; Value: Boolean); overload;
var
  lg: ILogGroup;
begin
  if Assigned(log_root) then
  begin
    log_root.find_or_add_group(group_name,lg);
    lg.AddLog(Kind,Text,Value);
  end else begin

  end;
end;

// Global AddLog with composed string
procedure AddComposedLog(const group_name: string; Kind: TLog_Kind; Const fmt: string; Const Params: Array of const); overload;
var
  lg: ILogGroup;
begin
  if Assigned(log_root) and log_root.find_or_add_group(group_name,lg) then
  begin
    lg.AddLog(Kind,fmt,Params);
  end else begin

  end;
end;

// add a log entry to a given log group
procedure AddLog(group: ILogGroup; Kind: TLog_Kind; Const Text: string); overload;
begin
  if Assigned(group) then
  begin
    group.AddLog(Kind,Text);
  end else
  if assigned(log_root) then
  begin
    log_root.AddLog(Kind,text);
  {$IFDEF DEBUG}
  end else begin
    DbgOut(Text);
  {$ENDIF}
  end;
end;

// same as above, with a formatted text
procedure AddLog(group: ILogGroup; Kind: TLog_Kind; Const fmt: string; Const Params: Array of const); overload;
begin
  if Assigned(group) then
  begin
    group.AddLog(Kind,fmt,Params);
  end else
  if assigned(log_root) then
  begin
    log_root.AddLog(Kind,fmt,Params);
  {$IFDEF DEBUG}
  end else begin
    try
      dbgout(fmt,params);
      except;
    end;
  {$ENDIF}
  end;
end;
// add a log entry for a switch value
procedure AddLog(group: ILogGroup; Kind: TLog_Kind; Const Text: string; Value: Boolean); overload;
begin
  if Assigned(group) then
  begin
    group.AddLog(Kind,Text,value);
  end else
  if assigned(log_root) then
  begin
    log_root.AddLog(Kind,text,value);
  {$IFDEF DEBUG}
  end else begin
    try
      dbgout(Text,[value]);
      except;
    end;
  {$ENDIF}
  end;
end;

function AddLogGroup(const name: string; group: ILogGroup): ILogGroup;
begin
  if assigned(group) then
  begin
     result := group.new_group(name);
  end else
  if assigned(log_root) then
  begin
    result := log_root.new_group(name);
  end;
end;


{ Tlog_kind_helper }

function Tlog_kind_helper.GetAsText: string;
begin
  case Self of
    LK_DECORATION: Result := '------------';
    LK_GROUP: Result := 'Group';
    LK_SHOWSTOPPER: Result := 'Show Stopper';
    LK_CRITICAL: Result := 'Critical';
    LK_EXCEPTION: Result := 'Exception';
    LK_ERROR: Result := 'Error';
    LK_WARNING: Result := 'Warning';
    LK_HINT: Result := 'Hint';
    LK_IMPORTANT: Result := 'Important';
    LK_INFO: Result := 'Info';
    LK_VERBOSE: Result := 'Verbose';
    LK_VERYVERBOSE: Result := 'Very Verbose';
    LK_DEBUG: Result := 'Debug';
  else
    Result := 'Unknown';
  end;
end;

procedure Tlog_kind_helper.SetAsKind(const Value: string);
var
  lk: Tlog_kind;
begin
  for lk := LK_DECORATION to LK_DEBUG do
    if SameText(lk.AsText, Value) then
    begin
      Self := lk;
      Exit
    end;
  self := LK_INFO;
end;

end.
